"""Qbis discovery models."""

from __future__ import annotations

from .const import (
    KEY_OUTPUT_ACTIONS,
    KEY_OUTPUT_ID,
    KEY_OUTPUT_LOCATION,
    KEY_OUTPUT_LOCATION_ID,
    KEY_OUTPUT_NAME,
    KEY_OUTPUT_PROPERTIES,
    KEY_OUTPUT_REF_ID,
    KEY_OUTPUT_TYPE,
)

KEY_DEVICES = "devices"

KEY_DEVICE_FUNCTIONBLOCKS = "functionBlocks"
KEY_DEVICE_ID = "id"
KEY_DEVICE_IP = "ip"
KEY_DEVICE_MAC = "mac"
KEY_DEVICE_NAME = "name"
KEY_DEVICE_SERIAL_NR = "serialNr"
KEY_DEVICE_TYPE = "type"
KEY_DEVICE_VERSION = "version"


class QbusMqttOutput:
    """MQTT representation of a Qbus output."""

    def __init__(self, data: dict, device: QbusMqttDevice) -> None:
        """Initialize based on a json loaded dictionary."""
        self._data = data
        self._device = device

    @property
    def id(self) -> str:
        """Return the id."""
        return self._data.get(KEY_OUTPUT_ID) or ""

    @property
    def type(self) -> str:
        """Return the type."""
        return self._data.get(KEY_OUTPUT_TYPE) or ""

    @property
    def name(self) -> str:
        """Return the name."""
        return self._data.get(KEY_OUTPUT_NAME) or ""

    @property
    def ref_id(self) -> str:
        """Return the ref id."""
        return self._data.get(KEY_OUTPUT_REF_ID) or ""

    @property
    def properties(self) -> dict:
        """Return the properties."""
        return self._data.get(KEY_OUTPUT_PROPERTIES) or {}

    @property
    def actions(self) -> dict:
        """Return the actions."""
        return self._data.get(KEY_OUTPUT_ACTIONS) or {}

    @property
    def location(self) -> str:
        """Return the location."""
        return self._data.get(KEY_OUTPUT_LOCATION) or ""

    @property
    def location_id(self) -> int:
        """Return the location id."""
        return self._data.get(KEY_OUTPUT_LOCATION_ID) or 0

    @property
    def device(self) -> QbusMqttDevice:
        """Return the device."""
        return self._device


class QbusMqttDevice:
    """MQTT representation of a Qbus device."""

    def __init__(self, data: dict) -> None:
        """Initialize based on a json loaded dictionary."""
        self._data = data

    @property
    def id(self) -> str:
        """Return the id."""
        return self._data.get(KEY_DEVICE_ID) or ""

    @property
    def ip(self) -> str:
        """Return the ip address."""
        return self._data.get(KEY_DEVICE_IP) or ""

    @property
    def mac(self) -> str:
        """Return the ip address."""
        return self._data.get(KEY_DEVICE_MAC) or ""

    @property
    def name(self) -> str:
        """Return the ip address."""
        return self._data.get(KEY_DEVICE_NAME) or ""

    @property
    def serial_number(self) -> str:
        """Return the serial number."""
        return self._data.get(KEY_DEVICE_SERIAL_NR) or ""

    @property
    def type(self) -> str:
        """Return the mac address."""
        return self._data.get(KEY_DEVICE_TYPE) or ""

    @property
    def version(self) -> str:
        """Return the version."""
        return self._data.get(KEY_DEVICE_VERSION) or ""

    @property
    def outputs(self) -> list[QbusMqttOutput]:
        """Return the outputs."""

        outputs: list[QbusMqttOutput] = []

        if self._data.get(KEY_DEVICE_FUNCTIONBLOCKS):
            outputs = [QbusMqttOutput(x, self) for x in self._data[KEY_DEVICE_FUNCTIONBLOCKS]]

        return outputs


class QbusDiscovery:
    """MQTT representation of a Qbus config."""

    def __init__(self, data: dict) -> None:
        """Initialize based on a json loaded dictionary."""
        if KEY_DEVICES in data:
            self._devices = [QbusMqttDevice(x) for x in data[KEY_DEVICES]]

        self._name = data["app"]

    def get_device_by_id(self, id: str) -> QbusMqttDevice | None:
        """Get the device by id."""
        return next((x for x in self.devices if x.id.casefold() == id.casefold()), None)

    def get_device_by_serial(self, serial: str) -> QbusMqttDevice | None:
        """Get the device by serial."""
        return next(
            (x for x in self.devices if x.serial_number.casefold() == serial.casefold()),
            None,
        )

    @property
    def devices(self) -> list[QbusMqttDevice]:
        """Return the devices."""
        return self._devices

    @property
    def name(self):
        """Return device name."""
        return self._name



"""Qbus MQTT API constants."""

TOPIC_PREFIX = "cloudapp/QBUSMQTTGW"

KEY_OUTPUT_ACTION = "action"
KEY_OUTPUT_ACTIONS = "actions"
KEY_OUTPUT_ID = "id"
KEY_OUTPUT_LOCATION = "location"
KEY_OUTPUT_LOCATION_ID = "locationId"
KEY_OUTPUT_NAME = "name"
KEY_OUTPUT_PROPERTIES = "properties"
KEY_OUTPUT_REF_ID = "refId"
KEY_OUTPUT_TYPE = "type"

KEY_PROPERTIES_AUTHKEY = "authKey"
KEY_PROPERTIES_SHUTTER_POSITION = "shutterPosition"
KEY_PROPERTIES_SLAT_POSITION = "slatPosition"
KEY_PROPERTIES_STATE = "state"
KEY_PROPERTIES_VALUE = "value"



"""Qbus MQTT factory."""

import json
import logging
from dataclasses import dataclass
from typing import Any, TypeVar

from .const import (
    KEY_PROPERTIES_AUTHKEY,
    TOPIC_PREFIX,
)
from .discovery import QbusDiscovery, QbusMqttDevice
from .state import (
    QbusMqttDeviceState,
    QbusMqttGatewayState,
    QbusMqttState,
    StateAction,
    StateType,
)

_LOGGER = logging.getLogger(__name__)


@dataclass
class QbusMqttRequestMessage:
    """Qbus MQTT request data class."""

    topic: str
    payload: str | bytes


class QbusMqttMessageFactory:
    """Factory methods for Qbus MQTT messages."""

    T = TypeVar("T", bound="QbusMqttState")

    def __init__(self) -> None:
        self._topic_factory = QbusMqttTopicFactory()

    def parse_gateway_state(self, payload: str | bytes) -> QbusMqttGatewayState | None:
        """Parse an MQTT message and return an instance
        of QbusMqttGatewayState if successful, otherwise None."""

        return self.deserialize(QbusMqttGatewayState, payload)

    def parse_discovery(self, payload: str | bytes) -> QbusDiscovery | None:
        """Parse an MQTT message and return an instance
        of QbusDiscovery if successful, otherwise None."""

        discovery: QbusDiscovery | None = self.deserialize(QbusDiscovery, payload)

        # Discovery data must include the Qbus device type and name.
        if discovery is not None and len(discovery.devices) == 0:
            _LOGGER.error("Incomplete discovery payload: %s", payload)
            return None

        return discovery

    def parse_device_state(self, payload: str | bytes) -> QbusMqttDeviceState | None:
        """Parse an MQTT message and return an instance
        of QbusMqttDeviceState if successful, otherwise None."""

        return self.deserialize(QbusMqttDeviceState, payload)

    def parse_output_state(self, cls: type[T], payload: str | bytes) -> T | None:
        """Parse an MQTT message and return an instance
        of T if successful, otherwise None."""

        return self.deserialize(cls, payload)

    def create_device_activate_request(self, device: QbusMqttDevice, prefix: str = TOPIC_PREFIX) -> QbusMqttRequestMessage:
        """Create a message to request device activation."""
        state = QbusMqttState(id=device.id, type=StateType.ACTION, action=StateAction.ACTIVATE)
        state.write_property(KEY_PROPERTIES_AUTHKEY, "ubielite")

        return QbusMqttRequestMessage(
            self._topic_factory.get_device_command_topic(device.id, prefix),
            self.serialize(state),
        )

    def create_device_state_request(self, device: QbusMqttDevice, prefix: str = TOPIC_PREFIX) -> QbusMqttRequestMessage:
        """Create a message to request a device state."""
        return QbusMqttRequestMessage(self._topic_factory.get_get_state_topic(prefix), json.dumps([device.id]))

    def create_state_request(self, ids: list[str], prefix: str = TOPIC_PREFIX) -> QbusMqttRequestMessage:
        """Create a message to request states."""
        return QbusMqttRequestMessage(self._topic_factory.get_get_state_topic(prefix), json.dumps(ids))

    def create_set_output_state_request(
        self, device: QbusMqttDevice, state: QbusMqttState, prefix: str = TOPIC_PREFIX
    ) -> QbusMqttRequestMessage:
        """Create a message to update the output state."""
        return QbusMqttRequestMessage(
            self._topic_factory.get_output_command_topic(device.id, state.id, prefix),
            self.serialize(state),
        )

    def serialize(self, obj: Any) -> str:
        """Convert an object to json payload."""
        return json.dumps(obj, cls=IgnoreNoneJsonEncoder)

    def deserialize(self, state_cls: type[Any], payload: str | bytes) -> Any | None:
        """Parse an MQTT message and return the requested type if successful, otherwise None."""

        if payload is None:
            _LOGGER.warning("Empty state payload for %s", state_cls.__name__)
            return None

        try:
            data = json.loads(payload)
        except ValueError:
            _LOGGER.error("Invalid state payload for %s: %s", state_cls.__name__, payload)
            return None

        return state_cls(data)


class QbusMqttTopicFactory:
    """Factory methods for topics of the Qbus MQTT API."""

    def get_gateway_state_topic(self, prefix: str = TOPIC_PREFIX) -> str:
        """Return the gateway state topic."""
        return f"{prefix}/state"

    def get_get_config_topic(self, prefix: str = TOPIC_PREFIX) -> str:
        """Return the getConfig topic."""
        return f"{prefix}/getConfig"

    def get_config_topic(self, prefix: str = TOPIC_PREFIX) -> str:
        """Return the config topic."""
        return f"{prefix}/config"

    def get_get_state_topic(self, prefix: str = TOPIC_PREFIX) -> str:
        """Return the getState topic."""
        return f"{prefix}/getState"

    def get_device_state_topic(self, device_id: str, prefix: str = TOPIC_PREFIX) -> str:
        """Return the state topic."""
        return f"{prefix}/{device_id}/state"

    def get_device_command_topic(self, device_id: str, prefix: str = TOPIC_PREFIX) -> str:
        """Return the 'set state' topic."""
        return f"{prefix}/{device_id}/setState"

    def get_output_command_topic(self, device_id: str, entity_id: str, prefix: str = TOPIC_PREFIX) -> str:
        """Return the 'set state' topic of an output."""
        return f"{prefix}/{device_id}/{entity_id}/setState"

    def get_output_state_topic(self, device_id: str, entity_id: str, prefix: str = TOPIC_PREFIX) -> str:
        """Return the state topic of an output."""
        return f"{prefix}/{device_id}/{entity_id}/state"


class IgnoreNoneJsonEncoder(json.JSONEncoder):
    """A json encoder to ignore None values when serializing."""

    def default(self, o):
        if hasattr(o, "__dict__"):
            # Filter out None values
            return {k: v for k, v in o.__dict__.items() if v is not None}
        return super().default(o)



from pathlib import Path

import setuptools

VERSION = "1.2.1"  # PEP-440

setuptools.setup(
    name="qbusmqttapi",
    version=VERSION,
    description="MQTT API for Qbus Home Automation.",
    url="https://github.com/Qbus-iot/qbusmqttapi",
    project_urls={
        "Source Code": "https://github.com/Qbus-iot/qbusmqttapi",
    },
    author="Koen Schockaert",
    author_email="ks@qbus.be",
    license="MIT License 2024",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Programming Language :: Python :: 3.10",
    ],
    package_data={'qbusmqttapi': ['py.typed']},
    python_requires=">=3.8",
    # Requirements
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
)



"""Qbus state models."""

from enum import StrEnum
from typing import Any

from .const import (
    KEY_OUTPUT_ACTION,
    KEY_OUTPUT_ID,
    KEY_OUTPUT_PROPERTIES,
    KEY_OUTPUT_TYPE,
    KEY_PROPERTIES_SHUTTER_POSITION,
    KEY_PROPERTIES_SLAT_POSITION,
    KEY_PROPERTIES_STATE,
    KEY_PROPERTIES_VALUE,
)

KEY_DEVICE_CONNECTABLE = "connectable"
KEY_DEVICE_CONNECTED = "connected"
KEY_DEVICE_ID = "id"
KEY_DEVICE_STATE_PROPERTIES = "properties"

KEY_GATEWAY_ID = "id"
KEY_GATEWAY_ONLINE = "online"
KEY_GATEWAY_REASON = "reason"


class StateType(StrEnum):
    """Values to be used as state type."""

    ACTION = "action"
    STATE = "state"


class StateAction(StrEnum):
    """Values to be used as state action."""

    ACTIVATE = "activate"
    ACTIVE = "active"


class QbusMqttGatewayState:
    """MQTT representation of a Qbus gateway state."""

    def __init__(self, data: dict) -> None:
        """Initialize based on a json loaded dictionary."""
        self.id: str | None = data.get(KEY_GATEWAY_ID)
        self.online: bool | None = data.get(KEY_GATEWAY_ONLINE)
        self.reason: str | None = data.get(KEY_GATEWAY_REASON)


class QbusMqttDeviceStateProperties:
    """MQTT representation of a Qbus device its state properties."""

    def __init__(self, data: dict) -> None:
        """Initialize based on a json loaded dictionary."""
        self.connectable: bool | None = data.get(KEY_DEVICE_CONNECTABLE)
        self.connected: bool | None = data.get(KEY_DEVICE_CONNECTED)


class QbusMqttDeviceState:
    """MQTT representation of a Qbus device state."""

    def __init__(self, data: dict) -> None:
        """Initialize based on a json loaded dictionary."""
        self.id: str | None = data.get(KEY_DEVICE_ID)

        properties = data.get(KEY_DEVICE_STATE_PROPERTIES)
        self.properties: QbusMqttDeviceStateProperties | None = (
            QbusMqttDeviceStateProperties(properties) if properties is not None else None
        )


class QbusMqttState:
    """MQTT representation of a Qbus state."""

    def __init__(
        self,
        data: dict | None = None,
        *,
        id: str | None = None,
        type: str | None = None,
        action: str | None = None,
    ) -> None:
        """Initialize state."""
        self.id: str = ""
        self.type: str = ""
        self.action: str | None = None
        self.properties: dict | None = None

        if data is not None:
            self.id = data.get(KEY_OUTPUT_ID, "")
            self.type = data.get(KEY_OUTPUT_TYPE, "")
            self.action = data.get(KEY_OUTPUT_ACTION)
            self.properties = data.get(KEY_OUTPUT_PROPERTIES)

        if id is not None:
            self.id = id

        if type is not None:
            self.type = type

        if action is not None:
            self.action = action

    def read_property(self, key: str, default: Any) -> Any:
        """Read a property."""
        return self.properties.get(key, default) if self.properties else default

    def write_property(self, key: str, value: Any) -> None:
        """Add or update a property."""
        if self.properties is None:
            self.properties = {}

        self.properties[key] = value


class QbusMqttOnOffState(QbusMqttState):
    """MQTT representation of a Qbus on/off output."""

    def __init__(
        self,
        data: dict | None = None,
        *,
        id: str | None = None,
        type: str | None = None,
    ) -> None:
        super().__init__(data, id=id, type=type)

    def read_value(self) -> bool:
        """Read the value of the Qbus output."""
        return self.read_property(KEY_PROPERTIES_VALUE, False)

    def write_value(self, value: bool) -> None:
        """Set the value of the Qbus output."""
        self.write_property(KEY_PROPERTIES_VALUE, value)


class QbusMqttAnalogState(QbusMqttState):
    """MQTT representation of a Qbus analog output."""

    def __init__(
        self,
        data: dict | None = None,
        *,
        id: str | None = None,
        type: str | None = None,
    ) -> None:
        super().__init__(data, id=id, type=type)

    def read_percentage(self) -> float:
        """Read the value of the Qbus output."""
        return self.read_property(KEY_PROPERTIES_VALUE, 0)

    def write_percentage(self, percentage: float) -> None:
        """Set the value of the Qbus output."""
        self.write_property(KEY_PROPERTIES_VALUE, percentage)


class QbusMqttShutterState(QbusMqttState):
    """MQTT representation of a Qbus shutter output."""

    def __init__(
        self,
        data: dict | None = None,
        *,
        id: str | None = None,
        type: str | None = None,
    ) -> None:
        super().__init__(data, id=id, type=type)

    def read_state(self) -> str | None:
        """Read the state of the Qbus output."""
        return self.read_property(KEY_PROPERTIES_STATE, None)

    def write_state(self, state: str) -> None:
        """Set the state of the Qbus output."""
        self.write_property(KEY_PROPERTIES_STATE, state)

    def read_position(self) -> int | None:
        """Read the position of the Qbus output."""
        return self.read_property(KEY_PROPERTIES_SHUTTER_POSITION, None)

    def write_position(self, percentage: int) -> None:
        """Set the position of the Qbus output."""
        self.write_property(KEY_PROPERTIES_SHUTTER_POSITION, percentage)

    def read_slat_position(self) -> int | None:
        """Read the slat position of the Qbus output."""
        return self.read_property(KEY_PROPERTIES_SLAT_POSITION, None)

    def write_slat_position(self, percentage: int) -> None:
        """Set the slat position of the Qbus output."""
        self.write_property(KEY_PROPERTIES_SLAT_POSITION, percentage)
